/* LOGMSG.C - Utility procedures to minimise stderr output */

#include <stdio.h>
#include "environ.h"

#ifdef OS9
#include <varargs.h>
#include <strings.h>
#endif

#ifdef MSDOS
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#endif

#ifdef LINUX
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#endif

/*
    Define the buffer structure
*/
struct logent {                        /* one buffer entry */
                  time_t timestamp;
                  char   msg[MAXLINE+1];
              };
struct logent _msgbuf[BUFFLIM];        /* the circular buffer */
int           _msgbufptr;              /* buffer update pointer */
time_t        _last_dump;              /* time of last dump */
int           _first_msg;              /* first message to dump */
int           _first_time  = 1;        /* initialisation control flag */
int           _runtype = INDIRECT_LOG; /* selects direct or indirect output */
int           _fileno = STDERR_LOG;    /* selects stdout or stderr */

#define  INCPTR(p) (p+1)%BUFFLIM

/*
    logsetup() - selects operational mode and output file no
*/
void logsetup(int r, int f)
{
    _runtype = r;
    _fileno  = f;
}

/*
    logmsg() - handles message logging.

    If the debugging flag DIRECT_LOG is on this call is equivalent
    to 'fprintf(stderr, format, ....)'. If it is off each message
    is timestamped and written to a circular buffer, from where it
    is printed by logdump().
*/
#ifdef OS9
void logmsg(fmt, va_alist)
char    *fmt;
va_dcl
#else
void logmsg(char* fmt, ...)
#endif
{
    va_list   apx;
    char      linebuff[1000];
    time_t    now;
    struct tm *tblock;
    int       i;
    int       lth;

    /*
        Set up the message to be logged and get the time
    */
    va_start(apx, fmt);
    lth = vsprintf(linebuff, fmt, apx);
    if (linebuff[lth-1] == '\n')
        linebuff[lth-1] = 0;
    va_end(apx);

    now = time(NULL);

    /*
        This code handles direct output
    */
    if (_runtype == DIRECT_LOG)
    {
        tblock = localtime(&now);
        if (_fileno == STDOUT_LOG)
            fprintf(stdout,
                    "%02d:%02d:%02d %s\n",
                    tblock->tm_hour,
                    tblock->tm_min,
                    tblock->tm_sec,
                    linebuff);
        else
            fprintf(stderr,
                    "%02d:%02d:%02d %s\n",
                    tblock->tm_hour,
                    tblock->tm_min,
                    tblock->tm_sec,
                    linebuff);
        return;
    }

    /*
        Indirect output.

        Initialise the buffer the first time through
    */
    if (_first_time)
    {
        for (i=0; i<BUFFLIM; i++)
            _msgbuf[i].timestamp = 0;
        _msgbufptr = 0;
        _last_dump = 0;
        _first_time = 0;
    }

    /*
        Add the message to the buffer.
    */
    _msgbuf[_msgbufptr].timestamp = now;
    strncpy(_msgbuf[_msgbufptr].msg, linebuff, MAXLINE);
    _msgbuf[_msgbufptr].msg[MAXLINE] = 0;

    /*
        Adjust the pointers. The storage pointer is incremented
        modulo the buffer size. If this catches up with the dump
        pointer this is carried along with it.
    */
    _msgbufptr = INCPTR(_msgbufptr);
    if (_msgbufptr == _first_msg)
        _first_msg = INCPTR(_msgbufptr);
}

/*
    logdump() - prints the buffer set up by logmsg().

    If the debugging flag DIRECT_LOG is on this function does nothing.
    If it is off all unprinted messages in the buffer are printed.
*/
void logdump(void)
{
    struct tm *tblock;
    int       i;
    int       last;

    /*
        Do nothing if nothing has been logged or the
        direct output flag is on
    */
    if (_runtype == DIRECT_LOG || _first_time)
        return;

    /*
        Scan the buffer, outputting all unprinted messages
        in order of arrival. Keep track of the last message
        output.
    */
    i = _first_msg;
    do {
        if (_msgbuf[i].timestamp > 0
            && _msgbuf[i].timestamp >= _last_dump) {
            tblock = localtime(&_msgbuf[i].timestamp);
            if (_fileno == STDOUT_LOG )
                fprintf(stdout,
                        "%02d:%02d:%02d %s\n",
                        tblock->tm_hour,
                        tblock->tm_min,
                        tblock->tm_sec,
                        _msgbuf[i].msg);
            else
                fprintf(stderr,
                        "%02d:%02d:%02d %s\n",
                        tblock->tm_hour,
                        tblock->tm_min,
                        tblock->tm_sec,
                        _msgbuf[i].msg);
            last = i;
        }
        i = INCPTR(i);
    }
    while (i != _msgbufptr);

    /*
        Reset the dump starting pointer and dump time.
    */
    _first_msg = INCPTR(last);
    _last_dump = time(NULL);
}
